# DISCLAIMER:  The Atm-I and calibration model coefficients specified for this demonstration
#              CMAC V1.1 were derived from image data acquired primarily over North America
#              between the months of June and August, when the solar zenith angle (SZA) deviated
#              less than 10 degrees relative to its minimum position around the summer solstice.
#              ARSI does not guarantee accuracy/reliability of CMAC v1.1 correction on Sentinel-2
#              images acquired with SZA greater than 30 degrees.
#
#              CMAC v1.1 does not correct for the interactive effects of Atm-I and 
#              SZA, which can be significant. Consideration of such effects and others
#              will be standard in future commercial versions of CMAC.
#
#              CMAC v1.1 is not approved for use with data other than that acquired by
#              Sentinel-2. Calibration of CMAC for other satellite imaging systems is in
#              progress. 

import argparse
import datetime as dt
import os
from pathlib import Path
import sys
import time

import numpy as np
from osgeo import gdal
from osgeo import gdalconst

import cmac_config as cfg
from cmac_correct import correct_toa_images
import cmac_gdal_io as io
from cmac_generate_raw_atmi import generate_raw_atmi
import cmac_load as load
from cmac_mask import create_nodata_mask_images
import cmac_metadata as meta
from cmac_postprocess_atmi import post_process_raw_atmi
from cmac_report import generate_summary_report
from cmac_resample import upsample_by_2, downsample_by_2
import cmac_utility as util



def main():

    # Initializations:
    cmac_overall_status = 0

    # Processing configuration info. Command-line invocation is
    # assumed as follows:
    # python path/to/PyCMACV11.py path/to<name of config file>
    parser = argparse.ArgumentParser()
    parser.add_argument('cfg_filename')
    cmd_line_args = parser.parse_args()
    cfg_filename = cmd_line_args.cfg_filename

    #cfg_filename = 'C:/Users/tarug/Desktop/cfg_test.json'
    #cfg_filename = 'D:/PyCMACV11/pycmacV11_cfg.json'
    cfg_param_dict = cfg.read_config(cfg_filename)
    if not cfg_param_dict:
        err_msg = 'ERROR: cannot read configuration file '  \
            + f'"{cfg_filename}"'
        print(err_msg)
        return -1

    # CMAC processing date (YYYYmmdd)
    cmac_proc_date = dt.datetime.strftime(dt.datetime.now(),
                                        '%Y%m%d')
    print(f'PROCESSING DATE:    {cmac_proc_date}')

    # Start timing CMAC processing.
    cmac_start_time = time.perf_counter()

    #################### CMAC INTIALIZATION ################################

    # Used to check whether Atm-I and/or CMAC corrected pixel values exceed
    # the maximum value for 16-bit unsigned integers.
    max_uint16_val = np.iinfo(np.uint16).max

    # Get configuration file info.
    user_id = cfg_param_dict["user_id"]
    product_id = cfg_param_dict["product_id"]
    file_ext = cfg_param_dict["file_ext"]

    # Construct the image paths.
    s2_image_basename = util.construct_s2_image_basename(product_id)
    s2_datatake_datetime \
        = util.get_s2_datatake_datetime_from_product_id(product_id)
    s2_tile_id = util.get_s2_tile_id_from_product_id(product_id)

    toa_root_drive = cfg_param_dict["root_toa_path"]
    cmac_root_drive = cfg_param_dict["root_cmac_path"]

    nodata_value = cfg_param_dict["nodata_value"]

    if not Path(toa_root_drive).exists():
        err_msg = 'ERROR: specified input TOA root path ' \
            + f'"{toa_root_drive}" not found'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # Get the final path to the TOA image data.
    parent_toa_path \
        = util.construct_parent_path(toa_root_drive, user_id,
                                     product_id, 'TOA', '')
    if not parent_toa_path:
        err_msg = f'ERROR: specified image path "{parent_toa_path}" ' \
            + 'not found'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    s2_image_data_path = util.get_toa_image_path(parent_toa_path)
    if s2_image_data_path == '':
        err_msg = 'ERROR: cannot determine final S2 TOA image path ' \
            + f'from "{parent_toa_path}'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # If the CMAC image path doesn't exist, create it.
    parent_cmac_path \
        = util.construct_parent_path(cmac_root_drive, user_id,
                                     product_id, 'CMAC', cmac_proc_date)
    if parent_cmac_path == '':
        err_msg = 'ERROR: could not create parent CMAC output path'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1
    path = Path(parent_cmac_path)
    if not path.exists():
        print(f'"{parent_cmac_path}" not found, creating...')
        path.mkdir(parents=True, exist_ok=True)
    else:
        print(f'"{parent_cmac_path}" found, clearing previous processing...')
        util.clean_cmac_directory(parent_cmac_path)

    # Construct the list of filenames for the TOAR image band(s) used
    # in Atm-I estimation.
    toa_bands_to_model_indices = cfg_param_dict["bands_to_model_indices"]
    num_bands_to_model = len(toa_bands_to_model_indices)
    toa_bands_to_model_list = util.get_band_ids_list(toa_bands_to_model_indices)
    if not toa_bands_to_model_list:
        err_msg = 'ERROR: cannot create list of band IDs for Atm-I modeling'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    toa_model_image_filenames_list \
        = util.construct_image_filenames(s2_image_data_path, s2_image_basename,
                                         '', '', toa_bands_to_model_list,
                                         file_ext)

    # Construct the list of filenames for the TOAR image band(s) to be
    # corrected.
    toa_bands_to_correct_indices = cfg_param_dict["bands_to_correct_indices"]
    num_bands_to_correct = len(toa_bands_to_correct_indices)
    toa_bands_to_correct_list = util.get_band_ids_list(toa_bands_to_correct_indices)
    if not toa_bands_to_correct_list:
        err_msg = 'ERROR: cannot create list of band IDs to be corrected'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    toa_correct_image_filenames_list \
        = util.construct_image_filenames(s2_image_data_path, s2_image_basename,
                                         '', '', toa_bands_to_correct_list,
                                         file_ext)

    # Construct the TOAR 10m and 20m reference filenames. The 10m reference
    # will come from the first band specified in toa_bands_to_correct_list,
    # and the 20m reference will come from the last band specified. For
    # typical CMAC correction these will most likely be the Blue (B02) and
    # NIR (B8A) bands, respectively.
    first_correct_band_index = 0
    last_correct_band_index = len(toa_bands_to_correct_indices) - 1
    band_id_30m = '';
    if (first_correct_band_index < 3) or (last_correct_band_index == 6):
        band_id_30m = toa_bands_to_correct_list[first_correct_band_index]
    if (last_correct_band_index >= 3) and (last_correct_band_index != 6):
        band_id_60m = toa_bands_to_correct_list[last_correct_band_index]

    toa_ref_filename_30m = s2_image_data_path + '/' + s2_image_basename \
        + '_' + band_id_30m + '.' + file_ext
    if not Path(toa_ref_filename_30m).exists():
        err_msg = 'ERROR: 10m reference band file not present'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # toa_ref_filename_60m = s2_image_data_path + '/' + s2_image_basename \
        # + '_' + band_id_60m + '.' + file_ext
    # if not Path(toa_ref_filename_60m).exists():
        # err_msg = 'ERROR: 20m reference band file not present'
        # print(err_msg)
        # cmac_stop_time = time.perf_counter()
        # elapsed_time = cmac_stop_time - cmac_start_time
        # print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        # return -1

    # Construct the list of basenames for the Atm-I image files.
    raw_atmi_image_basename = parent_cmac_path + '/' + 'CMAC_' \
        + s2_image_basename + '_raw_AtmIndex'
    smoothed_atmi_image_basename = parent_cmac_path + '/' + 'CMAC_' \
        + s2_image_basename + '_AtmIndex'

    # We'll need versions for the 10m and 20m Atm-I image files.
    raw_30m_atmi_GTiff_filename = raw_atmi_image_basename + '_30m.tif'
    raw_30m_atmi_product_filename = raw_atmi_image_basename + '_30m.' \
        + file_ext
    smoothed_30m_atmi_GTiff_filename = smoothed_atmi_image_basename \
        + '_30m.tif'
    smoothed_30m_atmi_product_filename = smoothed_atmi_image_basename \
        + '_30m.' + file_ext
    # smoothed_60m_atmi_GTiff_filename = smoothed_atmi_image_basename \
        # + '_60m.tif'
    # smoothed_60m_atmi_product_filename = smoothed_atmi_image_basename \
        # + '_60m.' + file_ext

    ######### END CMAC INITIALIZATION ######################################


    ######### RETRIEVE METADATA ############################################

    # Get the path/name of the processing metadata file, then retrieve the
    # processing metadata.
    proc_metadata_filename = util.get_metadata_filename(parent_toa_path, 'PROCESSING')
    if not proc_metadata_filename:
        err_msg = 'ERROR: could not construct processing metadata filename'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # Get the metadata for the Atm-I bands to model.
    proc_metadata_atmi \
        = meta.get_proc_metadata(proc_metadata_filename, toa_bands_to_model_list)
    proc_metadata_correct \
        = meta.get_proc_metadata(proc_metadata_filename, toa_bands_to_correct_list)
    

    if not (proc_metadata_atmi and proc_metadata_correct):
        err_msg = 'ERROR: could not retrieve processing metadata'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # In future versions of CMAC, a call to get the mean solar angle position
    # goes here (when an Atm-I model accounting for SZA and/or SAA effects is
    # developed. That call will not be made in this demo version.

    ######### END RETRIEVE METADATA #########################################


    ######### GDAL INITIALIZATION ###########################################

    # Make sure the required GDAL image drivers are installed. For the current
    # iteration of CMAC for Sentinel-2, these are the 'GeoTIFF' and 'JP2OpenJPEG'
    # drivers. We won't be using these driver objects directly in any processing --
    # we're just testing for their existence, and aborting further processing
    # if for some reason either one or both aren't found.
    #
    # NOTE: GDALAllRegister(), which initializes ALL of GDAL's installed
    #       image drivers, does not have to be called from the main function
    #       in Python -- it is automatically called when the 'gdal' module
    #       is imported. Nice change from C/C++, where it does have to be
    #       explicitly called before calling on any other GDAL functionality.
    print('Locating required image drivers...')
    test_GTiff_driver = gdal.GetDriverByName('GTiff')
    test_product_driver = gdal.GetDriverByName('JP2OpenJPEG')
    if not (test_GTiff_driver and test_product_driver):
        err_msg = 'ERROR: one or more GDAL image drivers required for CMAC ' \
            + 'processing not found'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1
    test_GTiff_driver = None
    test_product_driver = None

    ########## END GDAL INITIALIZATION #####################################


    ########## INITIALZE TOA IMAGE DATA ####################################

    # Get the geometric transformation and projection info:

    # 10m first.
    print('Retrieving 10m band geometric transform/projection information...')
    georef_dataset_30m = io.open_dataset(toa_ref_filename_30m, gdalconst.GA_ReadOnly)
    if not georef_dataset_30m:
        err_msg = 'ERROR: cannot open "{toa_ref_filename_30m}" with read-only ' \
            + 'access'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    geotransform_30m = io.get_geotransform_info(georef_dataset_30m)
    proj_info_30m = io.get_projection_info(georef_dataset_30m)
    georef_dataset_30m = None

    if not geotransform_30m:
        err_msg = 'ERROR: could not retrieve 10m geometric transform and/or ' \
            f'projection information from "{toa_ref_filename_30m}"'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # Then 20m.
    # print('Retrieving 20m band geometric transform/projection information...')
    # georef_dataset_60m = io.open_dataset(toa_ref_filename_60m, gdalconst.GA_ReadOnly)
    # if not georef_dataset_60m:
        # err_msg = 'ERROR: cannot open "{toa_ref_filename_60m}" with read-only ' \
            # + 'access'
        # print(err_msg)
        # cmac_stop_time = time.perf_counter()
        # elapsed_time = cmac_stop_time - cmac_start_time
        # print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        # return -1

    # geotransform_60m = io.get_geotransform_info(georef_dataset_60m)
    # proj_info_60m = io.get_projection_info(georef_dataset_60m)
    # georef_dataset_60m = None

    # if not geotransform_60m:
        # err_msg = 'ERROR: could not retrieve 20m geometric transform and/or ' \
            # f'projection information from "{toa_ref_filename_60m}"'
        # print(err_msg)
        # cmac_stop_time = time.perf_counter()
        # elapsed_time = cmac_stop_time - cmac_start_time
        # print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        # return -1

    ######### END INITIALIZE TOA IMAGE DATA ################################


    ######### LOAD Atm-I MODEL BANDS #######################################

    # If any 20m band is an Atm-I model input, resample it to 10m. Replace
    # the original filename in  the modeling list with the resampled filename.
    # WARNING: As currently implemented, if no 10m visible/NIR band is used
    # in the modeling, ALL of the specified 20m bands will be resampled to
    # 10m. If a 60m band is found in the list, the program will terminate.
    print('Loading image band(s) needed for Atm-I estimation...')
    resampled_toa_filenames_GTiff = []
    resampled_toa_filenames_product = []
    found_60m_atmi_band_flag = False

    for i in range(num_bands_to_model):

        # Get the spatial resolution for the current band from the processing
        # metadata info. The data for each band is specified as follows:
        # [a, b, c] where a=quantification value, b=spatial resolution (m), and
        # c=radiometric offset
        spatial_res = proc_metadata_atmi[toa_bands_to_model_list[i]][1]

        # if spatial_res > 20:
            # err_msg = f'ERROR: invalid band {toa_bands_to_model_list[i]} specified -- ' \
                # + f'spatial resolution is {spatial_res}'
            # print(err_msg)
            # cmac_stop_time = time.perf_counter()
            # elapsed_time = cmac_stop_time - cmac_start_time
            # print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
            # return -1
        # elif spatial_res == 20:
            # print(f'Resampling band {toa_bands_to_model_list[i]} TOAR image data ' \
                # + 'to 10m resolution...')
            # resampled_file_basename = parent_cmac_path + '/' \
                # + s2_image_basename + '_' + toa_bands_to_model_list[i] \
                # + '_resampled.'
            # resampled_filename_GTiff = resampled_file_basename + 'tif'
            # resampled_filename_product = resampled_file_basename + file_ext

            # # Resample the image.
            # status = upsample_by_2(toa_model_image_filenames_list[i],
                                   # geotransform_30m, proj_info_30m,
                                   # resampled_filename_GTiff,
                                   # resampled_filename_product)
            # if status < 0:
                # err_msg = 'ERROR: could create resampled TOAR image ' \
                    # + f'"{resampled_file_basename}"'
                # print(err_msg)
                # cmac_stop_time = time.perf_counter()
                # elapsed_time = cmac_stop_time - cmac_start_time
                # print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
                # return -1

            # # Add this resampled filename to the list.
            # resampled_toa_filenames_GTiff.append(resampled_filename_GTiff)
            # resampled_toa_filenames_product.append(resampled_filename_product)

            # # Indicate we found a 20m band in the modeling data.
            # found_60m_atmi_band_flag = True

        # else:
        # 10m band -- just copy the filename to the list.
        resampled_toa_filenames_product.append(toa_model_image_filenames_list[i])

    # Make sure to use RESAMPLED image data for any 20m bands used in the modeling.
    toa_atmi_model_datasets = load.load_toa_images(resampled_toa_filenames_product,
                                                   toa_bands_to_model_list)
    if not toa_atmi_model_datasets:
        err_msg = 'ERROR: failed to load ATMI model inputs'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    ######### END LOAD Atm-I MODEL BANDS ###################################


    ######### GENERATE Atm-I NODATA MASKS ##################################

    print('Generating "nodata" mask images for Atm-I input bands...')
    nodata_mask_filenames_GTiff_atmi \
        = util.construct_image_filenames(parent_cmac_path, s2_image_basename,
                                         "", "nodata_mask", toa_bands_to_model_list,
                                         "tif")
    if not nodata_mask_filenames_GTiff_atmi:
        err_msg = 'ERROR: failed to generate Atm-I "nodata" mask image filenames'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        for ds in toa_atmi_model_datasets:
            ds = None
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    if file_ext != 'tif':
        nodata_mask_filenames_product_atmi \
            = util.construct_image_filenames(parent_cmac_path, s2_image_basename,
                                             "", "nodata_mask", toa_bands_to_model_list,
                                             file_ext)
        if not nodata_mask_filenames_product_atmi:
            err_msg = 'ERROR: failed to generate Atm-I "nodata" mask image filenames'
            print(err_msg)
            cmac_stop_time = time.perf_counter()
            for ds in toa_atmi_model_datasets:
                ds = None
            elapsed_time = cmac_stop_time - cmac_start_time
            print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
            return -1
    else:
        nodata_mask_filenames_product_atmi = nodata_mask_filenames_GTiff_atmi

    status = create_nodata_mask_images(toa_atmi_model_datasets,
                                       toa_bands_to_model_list,
                                       nodata_value,
                                       nodata_mask_filenames_GTiff_atmi,
                                       nodata_mask_filenames_product_atmi)
    if status < 0:
        err_msg = 'ERROR: cannot create required "nodata" mask images'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        for ds in toa_atmi_model_datasets:
            ds = None
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    ######### END GENERATE Atm-I NODATA MASKS ##############################


    ######### GENERATE RAW Atm-I ###########################################

    # Get the gridcell dimensions.
    atmi_gridcell_size = cfg_param_dict["atmi_gridcell_size"]

    print('Generating (raw) Atm-I image from VNIR band inputs...')
    print(f'Atm-I GRIDCELL SIZE (pixels):  {atmi_gridcell_size}')

    # Time how long it takes to create the final Atm-I image. This
    # will include generating the raw image and performing any/all
    # specified smoothing operations on it.
    atmi_start_time = time.perf_counter()

    # Create the raw Atm-I dataset.
    atmi_model_coeffs = cfg_param_dict["atmi_model_coeffs"]
    status = generate_raw_atmi(toa_atmi_model_datasets, toa_bands_to_model_list,
                               nodata_value, proc_metadata_atmi, atmi_gridcell_size,
                               atmi_model_coeffs, max_uint16_val, proj_info_30m,
                               geotransform_30m, nodata_mask_filenames_GTiff_atmi[0],
                               raw_30m_atmi_GTiff_filename, raw_30m_atmi_product_filename)

    # Free up resources by "closing" the open dataset(s).
    for ds in toa_atmi_model_datasets:
        ds = None

    if status < 0:
        err_msg = 'ERROR: could not create "raw" 10-m Atm-I dataset'
        print(err_msg)
        util.clean_cmac_directory(parent_cmac_path)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    ######### END GENERATE RAW Atm-I #######################################


    ######### SMOOTH RAW Atm-I IMAGE #######################################

    # Atm-I smoothing options.
    median_filter_on = cfg_param_dict["median_filter_on"]
    median_filter_radius = cfg_param_dict["median_filter_radius"]
    gaussian_filter_on = cfg_param_dict["gaussian_filter_on"]
    gaussian_filter_radius = cfg_param_dict["gaussian_filter_radius"]

    status = post_process_raw_atmi(raw_30m_atmi_GTiff_filename,
                                   nodata_mask_filenames_GTiff_atmi[0],
                                   atmi_gridcell_size[0], atmi_gridcell_size[1],
                                   nodata_value, median_filter_on,
                                   median_filter_radius, gaussian_filter_on,
                                   gaussian_filter_radius, smoothed_30m_atmi_GTiff_filename,
                                   smoothed_30m_atmi_product_filename,
                                   geotransform_30m, proj_info_30m)
    if status < 0:
        err_msg = 'error: generation of "smoothed" atm-i datasets failed'
        print(err_msg)
        util.clean_cmac_directory(parent_cmac_path)
        cmac_stop_time = time.per_counter()
        print(f'total processing time:  {elapsed_time:.2f} seconds')
        return -1

    atmi_stop_time = time.perf_counter()
    elapsed_atmi_time = atmi_stop_time - atmi_start_time
    print(f'TOTAL Atm-I PROCESSING TIME:  {elapsed_atmi_time:.2f} seconds')

    ######### END SMOOTH RAW Atm-I IMAGE ###################################


    ######### CORRECT TOA IMAGE BANDS ######################################

    # Correct all specified bands. Save the corrected images as part of the
    # CMAC product output.
    print('Initializing  CMAC correction processing:')

    # Initializations:

    # Construct the list of filenames for the image band(s) undergoing
    # CMAC correction. We'll need two sets of names -- one for the
    # GeoTiff "source" version of the output and one for the JP2
    # "product" version of the output.
    cmac_GTiff_correct_filenames_list \
        = util.construct_image_filenames(parent_cmac_path, s2_image_basename,
                                         'CMAC', 'sr', toa_bands_to_correct_list,
                                         'tif')
    cmac_product_correct_filenames_list \
        = util.construct_image_filenames(parent_cmac_path, s2_image_basename,
                                         'CMAC', 'sr', toa_bands_to_correct_list,
                                         file_ext)
    if not (cmac_GTiff_correct_filenames_list) \
        and (cmac_product_correct_filenames_list):
        err_msg = 'ERROR: could not construct corrected output filename lists'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # Calibration slope and offset models for the band(s) to correct.
    band_cal_slope_models_list = cfg_param_dict["cal_slope_models"]
    band_cal_offset_models_list = cfg_param_dict["cal_offset_models"]

    # Atm-I "capping" and "hard-stops".
    atmi_cap_on = cfg_param_dict["atmi_cap_on"]
    atmi_cap_level = cfg_param_dict["atmi_cap_level"]
    
    # Load the TOA band(s) to correct.
    toa_correct_datasets_list = load.load_toa_images(toa_correct_image_filenames_list,
                                                     toa_bands_to_correct_list)
    if not toa_correct_datasets_list:
        err_msg = 'ERROR: could not construct corrected output filename lists'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # Generate 'nodata' mask images for the band(s) to correct.
    print('Generating "nodata" mask image images for input band(s) to ' \
        + 'correct...')
    nodata_mask_filenames_GTiff_correct \
        = util.construct_image_filenames(parent_cmac_path, s2_image_basename,
                                         "", "nodata_mask", toa_bands_to_correct_list,
                                         "tif")
    if file_ext != 'tif':
        nodata_mask_filenames_product_correct \
            = util.construct_image_filenames(parent_cmac_path, s2_image_basename,
                                             "", "nodata_mask", toa_bands_to_correct_list,
                                             file_ext)
    else:
        nodata_mask_filenames_product_correct \
            = nodata_mask_filenames_GTiff_correct
    if not (nodata_mask_filenames_GTiff_correct \
        and nodata_mask_filenames_product_correct):
        err_msg = 'ERROR: cannot create output mask image filenames'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        for ds in toa_correct_datasets_list:
            ds = None
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    status \
        = create_nodata_mask_images(toa_correct_datasets_list,
                                    toa_bands_to_correct_list,
                                    nodata_value,
                                    nodata_mask_filenames_GTiff_correct,
                                    nodata_mask_filenames_product_correct)
    if status < 0:
        err_msg = 'ERROR: cannot create all required "nodata" ' \
            + 'mask images'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        for ds in toa_correct_datasets_list:
            ds = None
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # Load the mask images.
    nodata_mask_correct_datasets_list = []
    for fname in nodata_mask_filenames_product_correct:
        curr_ds = io.open_dataset(fname, gdalconst.GA_ReadOnly)
        nodata_mask_correct_datasets_list.append(curr_ds)

    if len(nodata_mask_correct_datasets_list) != num_bands_to_correct:
        err_msg = f'ERROR: {len(nodata_mask_correct_datasets_list)} mask ' \
            + f'loaded, but {num_bands_to_correct} mask bands are required'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        for ds in toa_correct_datasets_list:
            ds = None
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1
    
    # If one or more 20m bands needs correcting, generate a 20m
    # smoothed ATMI image. Break after the first 20m TOA image
    # found in the band list to correct.
    # print('Checking for 20m band(s) to correct...')
    # for i in range(num_bands_to_correct):
        # correct_band_meta \
            # = proc_metadata_correct[toa_bands_to_correct_list[i]]
        # if correct_band_meta[1] == 20:
            # print('Resampling 10m smoothed Atm-I image to 20m resolution...')
            # status = downsample_by_2(smoothed_30m_atmi_product_filename,
                                     # geotransform_60m, proj_info_60m,
                                     # smoothed_60m_atmi_GTiff_filename,
                                     # smoothed_60m_atmi_product_filename)
            # if status < 0:
                # print('ERROR: could not resample 10m smoothed Atm-I ' \
                    # + f'image "{smoothed_30m_atmi_product_filename}" ' \
                    # + '20m resolution')
                # cmac_stop_time = time.perf_counter()
                # for i in range(num_bands_to_correct):
                    # toa_correct_datasets_list[i] = None
                    # nodata_mask_correct_datasets_list[i] = None
                # elapsed_time = cmac_stop_time - cmac_start_time
                # print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
                # return -1

            # break

    # Load up the 10m and 20m (smoothed) Atm-I images.
    print('Loading "smoothed" 10m Atm-I image (and 20m if needed)...')
    smoothed_30m_atmi_dataset = io.open_dataset(smoothed_30m_atmi_product_filename,
                                                gdalconst.GA_ReadOnly)
    if not smoothed_30m_atmi_dataset:
        err_msg = 'ERROR: cannot load 10m Atm-I dataset ' \
            + f'"{smoothed_30m_atmi_product_filename}"'
        print(err_msg)
        cmac_stop_time = time.perf_counter()
        for i in range(num_bands_to_correct):
            toa_correct_datasets_list[i] = None
            nodata_mask_correct_datasets_list[i] = None
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    # if Path(smoothed_60m_atmi_product_filename).exists():
        # smoothed_60m_atmi_dataset \
            # = io.open_dataset(smoothed_60m_atmi_product_filename,
                              # gdalconst.GA_ReadOnly)
        # if not smoothed_60m_atmi_dataset:
            # err_msg = 'ERROR: cannot load 20m Atm-I dataset ' \
                # + f'"{smoothed_60m_atmi_product_filename}"'
            # print(err_msg)
            # cmac_stop_time = time.perf_counter()
            # for i in range(num_bands_to_correct):
                # toa_correct_datasets_list[i] = None
                # nodata_mask_correct_datasets_list[i] = None
            # smoothed_30m_atmi_dataset = None
            # elapsed_time = cmac_stop_time - cmac_start_time
            # print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
            # return -1

    # Time the correction step.
    correct_start_time = time.perf_counter()
    correct_stop_time = correct_start_time

    cal_slope_models_list = cfg_param_dict["cal_slope_models"]
    cal_offset_models_list = cfg_param_dict["cal_offset_models"]

    print('Correcting TOA images...')
    cmac_negative_correct_errors_list, cmac_below_one_correct_errors_list, \
        cmac_above_10k_correct_errors_list \
        = correct_toa_images(toa_correct_datasets_list, smoothed_30m_atmi_dataset,
                             nodata_mask_correct_datasets_list,
                             proc_metadata_correct, toa_bands_to_correct_list,
                             atmi_cap_on, atmi_cap_level, nodata_value,
                             cal_slope_models_list, cal_offset_models_list,
                             max_uint16_val, geotransform_30m,
                             proj_info_30m, cmac_GTiff_correct_filenames_list,
                             cmac_product_correct_filenames_list)

    # Close out the datasets. Has to be done to ensure they can be read in
    # the future.
    for i in range(num_bands_to_correct):
        toa_correct_datasets_list[i] = None
        nodata_mask_correct_datasets_list[i] = None
    smoothed_30m_atmi_dataset = None
    # if smoothed_60m_atmi_dataset is not None:
        # smoothed_60m_atmi_datast = None

    # Stop CMAC correction process timing.
    correct_stop_time = time.perf_counter()
    elapsed_correct_time = correct_stop_time - correct_start_time

    if len(cmac_negative_correct_errors_list) != num_bands_to_correct \
        or len(cmac_below_one_correct_errors_list) != num_bands_to_correct \
        or len(cmac_above_10k_correct_errors_list) != num_bands_to_correct:
        err_msg = 'ERROR: missing one or more correction error summary ' \
            + 'information checks'
        print(err_msg)
        elapsed_time = cmac_stop_time - cmac_start_time
        print(f'TOTAL PROCESSING TIME: {elapsed_time:.2f} seconds')
        return -1

    #print('STOP')

    ######### END CORRECT TOA IMAGE BANDS ##################################


    ######### GENERATE FINAL SUMMARY REPORT ################################

    # Stop processing timing -- generating the final report shouldn't really
    # count here.
    cmac_stop_time = time.perf_counter()
    elapsed_proc_time = cmac_stop_time - cmac_start_time
    print(f'TOTAL PROCESSING TIME: {elapsed_proc_time:.2f} seconds')

    # Generate the final report.
    print('Generating processing summary report...')
    cmac_cal_version = cfg_param_dict["cmac_cal_version"]
    cmac_software_version = cfg_param_dict["cmac_software_version"]
    status = generate_summary_report(smoothed_30m_atmi_product_filename,
                                     nodata_value, toa_bands_to_correct_list,
                                     elapsed_atmi_time, elapsed_correct_time,
                                     elapsed_proc_time,
                                     cmac_negative_correct_errors_list,
                                     cmac_below_one_correct_errors_list,
                                     cmac_above_10k_correct_errors_list,
                                     parent_cmac_path, s2_image_basename,
                                     cmac_proc_date, cmac_cal_version,
                                     cmac_software_version)
    if status < 0:
        err_msg = 'ERROR: could not generate summary report'
        print(err_msg)
        cmac_overall_status = -1

    ######### END GENERATE FINAL SUMMARY REPORT ############################

    # Clean up the GeoTiff "source" files -- they take up a lot of space.
    print('Performing final cleanup...')
    #util.clean_cmac_gtiffs(parent_cmac_path)
 
    
    return cmac_overall_status





if __name__ == '__main__':
    main()