import json
from pathlib import Path


# DISCLAIMER:  The Atm-I and calibration model coefficients specified below were derived
#              from image data acquired primarily over North America between the months
#              of June and August, when the solar zenith angle (SZA) deviated less than
#              10 degrees relative to its minimum position around the summer solstice.
#              ARSI does not guarantee accuracy/reliability of CMAC v1.1 correction on
#              Sentinel-2 images acquired with SZA greater than 30 degrees.
#
#              CMAC v1.1 does not correct for the interactive effects of Atm-I and 
#              SZA, which can be significant. Consideration of such effects and others
#              will be standard in future commercial versions of CMAC.
#
#              CMAC v1.1 is not approved for use with data other than that acquired by
#              Sentinel-2. Calibration of CMAC for other satellite imaging systems is in
#              progress. 


def read_config(cfg_filename):

    # Initializations go here.
    cfg_param_dict = {}
    err_flag = False
    err_msg = ''

    # Get the actual user-supplied info. The rest of
    # the configuration info, at least for CMAC V1.1
    # is "hard-coded"; future versions of this software
    # could add these to the configuration file (as in
    # earlier CMAC development).
    if not Path(cfg_filename).exists():
        err_msg = f'ERROR: configuration file "{cfg_filename}" not found'
        print(err_msg)
        return {}

    # Read the config file.
    with open(cfg_filename, 'r') as fid:
        json_cfg_data = json.load(fid)

    # Get the path information.
    root_toa_path = json_cfg_data["rootTOAFolder"]
    root_cmac_path = json_cfg_data["rootCMACFolder"]
    user_id = json_cfg_data["userID"]
    product_id = json_cfg_data["productID"]

    # Populate the configuration parameter data structure
    cfg_param_dict.update({"root_toa_path" : root_toa_path})
    cfg_param_dict.update({"root_cmac_path" : root_cmac_path})
    cfg_param_dict.update({"user_id" : user_id})
    cfg_param_dict.update({"product_id" : product_id})

    # Set the "demo" version parameters.
    cfg_param_dict.update({"cmac_cal_version" : "VNIR_Cal_20230218_V1B"})
    cfg_param_dict.update({"cmac_software_version" : "PyCMAC_Custom_AtmI"})
    cfg_param_dict.update({"file_ext" : "tif"})

    cfg_param_dict.update({"nodata_value" : 0})
    cfg_param_dict.update({"bands_to_model_indices" : [0, 2]})
    cfg_param_dict.update({"bands_to_correct_indices" : [0, 1, 2, 7]})

    cfg_param_dict.update({"atmi_cap_on" : True})
    cfg_param_dict.update({"atmi_cap_level" : 2000})
    cfg_param_dict.update({"atmi_model_coeffs" : []})
    cfg_param_dict.update({"atmi_gridcell_size" : [10, 10]})

    cfg_param_dict.update({"median_filter_on" : True})
    cfg_param_dict.update({"median_filter_radius" : 3})
    cfg_param_dict.update({"gaussian_filter_on" : True})
    cfg_param_dict.update({"gaussian_filter_radius" : 7})

    cfg_param_dict.update({"cal_offset_models" : [[], \
        [], [], []]})
    cfg_param_dict.update({"cal_slope_models" : [[], \
        [], [], []]})

    return cfg_param_dict








