import os
from pathlib import Path

from osgeo import gdal, gdalconst, ogr, osr



# Open an existing dataset. 'modeFlag' options
# should be gdalconst.GA_ReadOnly or gdalConst.GA_Update
def open_dataset(dataset_filename, modeFlag):

    image_dataset = gdal.Open(dataset_filename, modeFlag)
    if image_dataset is None or not image_dataset:
        err_msg = f'ERROR: cannot load {dataset_filename}'
        print(err_msg)
        return None

    return image_dataset



def get_raster_rows(image_dataset):
    return image_dataset.RasterYSize



def get_raster_cols(image_dataset):
    return image_dataset.RasterXSize



def get_raster_band_from_dataset(image_dataset):

    image_raster_band = image_dataset.GetRasterBand(1)
    if not image_raster_band:
        err_msg = 'ERROR: cannot access raster band from ' \
            + 'open dataset'
        print(err_msg)
        return None

    return image_raster_band



def get_projection_info(image_dataset):

    projString = image_dataset.GetProjection()
    if not projString:
        err_msg = 'ERROR: projection info not found'
        print(err_msg)
        return None

    return projString



def get_geotransform_info(image_dataset):

    geotransform_info_info = image_dataset.GetGeoTransform()
    if not geotransform_info_info:
        err_msg = 'ERROR: cannot retrieve geometric transformation'
        print(err_msg)
        return None

    return geotransform_info_info



def read_image_from_open_dataset(image_dataset):

    return get_raster_band_from_dataset(image_dataset).ReadAsArray()




def read_image_from_file(dataset_filename):

    image_dataset = open_dataset(dataset_filename, gdalconst.GA_ReadOnly)
    if image_dataset is None or not image_dataset:
        err_msg = 'ERROR: cannot open {dataset_filename} to read'
        print(err_msg)
        return None

    # Populate the image data buffer, then close the dataset.
    image_data = read_image_from_open_dataset(image_dataset)
    image_dataset = None
    num_rows = image_data.shape[0]
    num_cols = image_data.shape[1]
    if num_rows == 0 and num_cols == 0:
        err_msg = 'ERROR: cannot read image data'
        print(err_msg)
        return None

    return image_data




def create_dataset(out_dataset_filename,
                   gdal_driverName, num_rows, num_cols,
                   geotransform_info, projection_string):

    # Initializations.
    

    # Get the specified driver.
    gdal_driver = gdal.GetDriverByName(gdal_driverName)
    if not gdal_driver or gdal_driver is None:
        err_msg = f'ERROR: cannot find driver named {gdal_driverName}'
        print(err_msg)
        return -1

    # Make sure it can support create and/or copy create.
    metadata = gdal_driver.GetMetadata()
    supports_create = False
    if metadata.get(gdalconst.DCAP_CREATE) == 'YES':
        supports_create = True

    if not supports_create:
        err_msg = f'ERROR: {gdal_driverName} does not support Create()'
        print(err_msg)
        return -1

    # Create an 'empty' dataset.
    out_dataset = gdal_driver.Create(out_dataset_filename, xsize=num_cols,
                                     ysize=num_rows, bands=1,
                                     eType=gdalconst.GDT_UInt16)
 
    if not out_dataset:
        err_msg = f'ERROR: cannot create dataset named {out_dataset_filename}'
        print(err_msg)
        return -1

    # Set the output projection and geometric transformation to the
    # specified inputs.
    out_dataset.SetGeoTransform(geotransform_info)
    out_srs = osr.SpatialReference()
    out_srs.ImportFromWkt(projection_string)
    out_dataset.SetProjection(out_srs.ExportToWkt())

    # Close the newly created dataset.
    out_dataset = None

    return 0




def copy_to_product_dataset(source_dataset_filename, product_dataset_filename,
                            product_driver_name):

    # Do CreateCopy() if the output filename represents a different
    # format type than the source filename.
    if product_dataset_filename != source_dataset_filename:

        # Make sure the specified product driver exists and supports
        # CreateCopy().
        productDriver = gdal.GetDriverByName(product_driver_name)
        if not productDriver:
            err_msg = f'ERROR: driver named {product_driver_name} not found'
            print(err_msg)
            return -1

        supports_create_copy = False
        driver_metadata = productDriver.GetMetadata()
        if driver_metadata.get(gdalconst.DCAP_CREATECOPY) == 'YES':
            supports_create_copy = True
        if not supports_create_copy:
            err_msg = 'ERROR: specified product driver does not support CreateCopy()'
            print(err_msg)
            return -1

        # Open the source dataset.
        source_dataset = open_dataset(source_dataset_filename, gdalconst.GA_ReadOnly)
        if not source_dataset:
            err_msg = 'ERROR: cannot create copy dataset -- source dataset not open'
            print(err_msg)
            return -1

        # Copy-create the product dataset from the source. Set options specific to JP2
        # relating to reversibility of compression and compression quality.
        product_dataset = productDriver.CreateCopy(product_dataset_filename,
                                                  source_dataset, strict=0,
                                                  options=["QUALITY=25", "REVERSIBLE=YES", \
                                                      "YCBCR420=NO"])
        source_dataset = None
        if not product_dataset:
            err_msg = f'ERROR: copy create of {product_dataset_filename} failed'
            print(err_msg)
            return -1
        product_dataset = None

    return 0




def write_image_to_geotiff(out_image_data, out_proj_info,
                           out_geotransform_info,
                           out_image_path,
                           out_image_basename,
                           out_dtype=gdalconst.GDT_UInt16):

    # Create the output location if it doesn't already exist.
    path = Path(out_image_path)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)

    out_image_filename = out_image_path + os.sep \
        + out_image_basename + '.tif'

    # Get the required dimensions.
    out_num_rows = out_image_data.shape[0]
    out_num_cols = out_image_data.shape[1]

    # Create the output dataset.
    status = create_dataset(out_image_filename, 'GTiff', out_num_rows,
                            out_num_cols, out_geotransform_info,
                            out_proj_info)
    if status < 0:
        err_msg = f'ERROR: cannot create "{out_image_filename}" dataset'
        print(err_msg)
        return -1

    # Write the image data to the dataset.
    out_dataset = open_dataset(out_image_filename, gdalconst.GA_Update)
    if not out_dataset:
        err_msg = f'ERROR: cannot update "{out_image_filename} dataset'
        print(err_msg)
        return -1
    out_dataset.GetRasterBand(1).WriteArray(out_image_data)

    # Close the dataset, we're done.
    out_dataset = None

    return 0




def write_image_to_jp2(src_image_filename, out_image_filename):

    # If the source ('GeoTIFF') image doesn't exist,
    # stop.
    if not Path(src_image_filename).exists():
        err_msg = f'ERROR: "{src_image_filename}" not found'
        print(err_msg)
        return -1

    # Split the total product output path into the parent and
    # basename. If the output path happens to be different than
    # the GeoTIFF output path (as implemented, they shouldn't be)
    # create the path.
    src_image_path = Path(src_image_filename).parent
    out_image_filename_path = Path(out_image_filename)
    out_image_path = out_image_filename_path.parent
    out_image_basename = out_image_filename_path.name.split('.')[0]
    
    if not Path(out_image_path).exists():
        msg = f'WARNING: "{out_image_path}" different than ' \
            + f'"{src_image_path}", creating new location'
        print(msg)
        Path(out_image_path).mkdir(parents=True, exist_ok=True)


    # Copy-create the JP2 dataset from the source dataset.
    status = copy_to_product_dataset(src_image_filename,
                                     out_image_filename,
                                     'JP2OpenJPEG')
    if status < 0:
        err_msg = f'ERROR: cannot create "{out_image_filename}"'
        print(err_msg)
    
    return status
    



def write_image_to_file(image_buffer, num_rows, num_cols,
                        geotransform_info, proj_info_string,
                        GTiff_dataset_filename,
                        product_dataset_filename):

    # Initialization(s).
    final_status = 0

    # Create the GeoTIFF file version first:

    # Split the official GeoTIFF image filename and the
    # parent path to it.
    GTiff_filename_path = Path(GTiff_dataset_filename)
    GTiff_dataset_path = GTiff_filename_path.parent
    GTiff_dataset_basename = GTiff_filename_path.name.split('.')[0]

    status = write_image_to_geotiff(image_buffer, proj_info_string,
                                    geotransform_info, str(GTiff_dataset_path),
                                    GTiff_dataset_basename)
    if status < 0:
        err_msg = f'ERROR: cannot save to GeoTIFF file "{GTiff_dataset_filename}"'
        print(err_msg)
        return -1

    # If the output "product" is JP2, copy-create that file.
    if GTiff_dataset_filename != product_dataset_filename:

        status = write_image_to_jp2(GTiff_dataset_filename,
                                    product_dataset_filename)
        if status < 0:
            err_msg = f'ERROR: cannot save to JP2 file "{product_dataset_filename}"'
            print(err_msg)
            final_status = -1

    return final_status














