import numpy as np

import cmac_gdal_io as io
import cmac_metadata as meta
from cmac_gridcell_stats import get_gridcell_minimum
import cmac_utility



def generate_raw_atmi(atmi_model_datasets, atmi_bands_to_model_list,
                      nodata_value, proc_metadata_dict, atmi_gridcell_size,
                      atmi_model_coeffs, max_uint16_val, proj_info_string,
                      geotransform_info, nodata_mask_filename_GTiff,
                      GTiff_raw_atmi_filename, product_raw_atmi_filename):

    # Initializations.
    num_gridcell_rows = atmi_gridcell_size[0]
    num_gridcell_cols = atmi_gridcell_size[1]

    num_atmi_bands_to_model = len(atmi_bands_to_model_list)
    num_atmi_model_coeffs = num_atmi_bands_to_model + 1

    # Set up access to the (modeling) TOAR image data.

    num_toa_raster_bands_open = 0
    atmi_model_raster_buffers_list = []
    for dataset in atmi_model_datasets:
        curr_buff = io.read_image_from_open_dataset(dataset)
        if curr_buff is not None:
            atmi_model_raster_buffers_list.append(curr_buff)
            num_toa_raster_bands_open += 1
 
    if num_toa_raster_bands_open != num_atmi_bands_to_model:
        err_msg = f'ERROR: {num_toa_raster_bands_open} raster bands ' \
            + 'returned, but {num_atmi_bands_to_model} raster bands ' \
            + 'required'
        print(err_msg)
        return -1

    # Allocate the set of gridcell data buffers for the modeling bands.
    band_toa_gridcell_buffers_list = []
    for i in range(num_atmi_bands_to_model):
        curr_gridcell_buff = np.zeros((num_gridcell_rows, num_gridcell_cols),
                                      dtype=np.uint16)
        band_toa_gridcell_buffers_list.append(curr_gridcell_buff);

    # Allocate the output gridcell data buffer.
    out_atmi_gridcell_buff = np.zeros((num_gridcell_rows, num_gridcell_cols),
                                      dtype=np.uint16)

    # Get the radiometric offsets for each band.
    band_radoffsets_list = []
    for i in range(num_atmi_bands_to_model):
        band_id = atmi_bands_to_model_list[i]
        band_radoffsets_list.append(proc_metadata_dict[band_id][2])

    # Allocate full-sized image buffers for the 'raw' Atm-I data and
    # data mask.
    num_image_rows = io.get_raster_rows(atmi_model_datasets[0])
    num_image_cols = io.get_raster_cols(atmi_model_datasets[0])
    
    print('Allocating full-sized Atm-I and "nodata" mask buffers...')
    out_atmi_image_buff = np.zeros((num_image_rows, num_image_cols))

    # MAIN GRIDCELL PROCESSING LOOP
    rw_height = rw_width = 0
    
    # Flag indicating we estimated an Atm-I level beyond the
    # range of a 16-bit unsigned integer value.
    atmi_above_uint16_maxval = False

    for i in range(0, num_image_rows, num_gridcell_rows):
        if (i % 100) == 0:
            print(f'Processing image row {i+1} ...' + '\r')

        if (i + num_gridcell_rows) > num_image_rows:
            rw_height = num_image_rows - i
        else:
            rw_height = num_gridcell_rows

        for j in range(0, num_image_cols, num_gridcell_cols):

            if (j + num_gridcell_cols) > num_image_cols:
                rw_width = num_image_cols - j
            else:
                rw_width = num_gridcell_cols

            # Read each band's gridcell buffer here.
            atmi_input_vec = []

            for r in range(0, num_toa_raster_bands_open):
                band_image = atmi_model_raster_buffers_list[r]
                buff = band_image[i:i+rw_height, j:j+rw_width]

                # Apply the radiometric offset.
                rad_offset = band_radoffsets_list[r]
                buff_nodata = (buff == nodata_value)
                buff = buff + rad_offset
                buff[buff_nodata] = nodata_value

                # Get the gridcell minimum for this band.
                min_val = get_gridcell_minimum(buff, nodata_value)

                # Add this to the model input list.
                atmi_input_vec.append(min_val)

            # Add 1.0 to the list (for the constant term in the model).
            atmi_input_vec.append(1.0)

            # Determine the ATMI level.
            log_atmi_level = 0.0
            for n in range(0, num_atmi_model_coeffs):
                log_atmi_level += atmi_model_coeffs[n] * atmi_input_vec[n]

            # With the use of "hard" Atm-I stop limits, pixels with Atm-I levels
            # this low will not be corrected. These should be considered "errors"
            # in Atm-I estimation, suggesting issue(s) with the Atm-I model itself.
            # They should occur rarely, if ever, perhaps only if the pixel value itself
            # is bad due to a data drop-type event.
            if log_atmi_level < 0.0:
                atmi_level = 0
            elif log_atmi_level < 1.0:
                atmi_level = 1
            else:
                atmi_level = np.round(np.exp(log_atmi_level)).astype(np.int)

            # Make sure there isn't an Atm-I level above the maximum range of
            # 16-bit unsigned integer values. "clamp" to this level if any
            # such level is found (should happen rarely, if ever).
            # Issue a warning if this occurs after processing in this loop
            # has finished.
            if atmi_level > max_uint16_val:
                atmi_above_uint16_maxval = True
                atmi_level = max_uint16_val
            else:
                atmi_level = atmi_level.astype(np.uint16)

            # Generate the output Atm-I gridcell.
            out_atmi_gridcell_buff = atmi_level

            # Write the gridcell to the final raw Atm-I image buffer.
            out_atmi_image_buff[i:i+rw_height, j:j+rw_width] \
                = atmi_level
            #out_atmi_image_buff[i:i+rw_height, j:j+rw_width] \
            #    = atmi_level

    # END MAIN GRIDCELL PROCESSING LOOP
    if atmi_above_uint16_maxval:
        warn_msg = 'WARNING: Atm-I level estimated beyond the range of 16-bit ' \
            + 'unsigned integer values for at least one gridcell. These have ' \
            + 'been clamped to the maximum value for future processing'
        print(warn_msg)
    print('')


    # MASK PROCESSING

    # Read in the nodata mask image to use.
    nodata_mask = io.read_image_from_file(nodata_mask_filename_GTiff)

    # Apply the mask.
    print('Applying "nodata" mask to "raw" Atm-I image data...')
    out_atmi_image_buff = out_atmi_image_buff * nodata_mask

    # Save the masked image to output.
    print('Saving "masked" Atm-I image raster(s)...')
    status = io.write_image_to_file(out_atmi_image_buff, num_image_rows,
                                    num_image_cols, geotransform_info,
                                    proj_info_string, GTiff_raw_atmi_filename,
                                    product_raw_atmi_filename)
    if status < 0:
        err_msg = f'ERROR: cannot create "raw" Atm-I GeoTiff and/or product ' \
            + 'images'
        print(err_msg)

    return status

