from decimal import Decimal

from bs4 import BeautifulSoup
import numpy as np


# Global variables.
image_band_list = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06',  \
    'B07', 'B08', 'B8A', 'B09', 'B10', 'B11', 'B12']

# Map a band string to an index into an S2proc_metadatadata
# data structure member containing information for ALL of the
# 10m/20m image bands. The band order is assumed/expected to
# be from shortest wavelength to longest.
def map_imageband_to_band_index(img_band_id):

    band_index = -1000;

    if img_band_id == 'B02':
        band_index = 0
    elif img_band_id == 'B03':
        band_index = 1
    elif img_band_id == 'B04':
        band_index = 2
    elif img_band_id == 'B05':
        band_index = 3
    elif img_band_id == 'B06':
        band_index = 4
    elif img_band_id == 'B07':
        band_index = 5
    elif img_band_id == 'B08':
        band_index = 6
    elif img_band_id == 'B8A':
        band_index = 7
    elif img_band_id == 'B11':
        band_index = 8
    elif img_band_id == 'B12':
        band_index = 9

    if band_index < 0 or band_index > 9:
        err_msg = f'ERROR: band {img_band_id} index not found'
        print(err_msg)
        return -1

    return band_index


# Reverse mapping function. The index is the index position
# in the "master" band list
def map_band_index_to_imageband(band_index):
    
    band_id = ''

    if band_index == 0:
        band_id = 'B02'
    elif band_index == 1:
        band_id = 'B03'
    elif band_index == 2:
        band_id = 'B04'
    elif band_index == 3:
        band_id = 'B05'
    elif band_index == 4:
        band_id = 'B06'
    elif band_index == 5:
        band_id = 'B07'
    elif band_index == 6:
        band_id = 'B08'
    elif band_index == 7:
        band_id = 'B8A'
    elif band_index == 8:
        band_id = 'B11'
    elif band_index == 9:
        band_id = 'B12'
    else:
        err_msg = 'ERROR: no band ID corresponding to index ' \
            + f'{band_index} found'
        print(err_msg)
    
    return band_id




# Map an image band string to the corresponding ID band attribute
# string used in the Sentinel-2 product metadata. Primarily needed
# to obtain radiometric offset information in images processed
# under baseline version 4.00 or later (starting January 2022).
def map_imageband_to_metadataband_id_attribute(bands_to_process):

    metadata_band_attrs = []

    for band in bands_to_process:
        if band == 'B01':
            metadata_band_attrs.append('0')
        elif band == 'B02':
            metadata_band_attrs.append('1')
        elif band == 'B03':
            metadata_band_attrs.append('2')
        elif band == 'B04':
            metadata_band_attrs.append('3')
        elif band == 'B05':
            metadata_band_attrs.append('4')
        elif band == 'B06':
            metadata_band_attrs.append('5')
        elif band == 'B07':
           metadata_band_attrs.append('6')
        elif band == 'B08':
            metadata_band_attrs.append('7')
        elif band == 'B8A':
            metadata_band_attrs.append('8')
        elif band == 'B09':
            metadata_band_attrs.append('9')
        elif band == 'B10':
            metadata_band_attrs.append('10')
        elif band == 'B11':
            metadata_band_attrs.append('11')
        elif band == 'B12':
            metadata_band_attrs.append('12')
        else:
            print(f'60m band {band} encountered, skipping')

    return metadata_band_attrs




# BeautifulSoup method of locating data in an XML
# tree through a label.
def get_bsoup_data(bsoup_data, label):
    label_data = bsoup_data.find(label)
    return label_data



# Read from TOAR product's 'MTD_TL.xml' metadata file.
def get_mean_solar_angles(tile_metadata_filename):

    # Open the file and read in the contents.
    with open(tile_metadata_filename, 'r') as fid:
        metadata = fid.read()
        bsoup_metadata = BeautifulSoup(metadata, "xml")

    # Navigate the metadata "tree" to the mean solar zenith and
    # azimuth angle declarations.
    zenith_angle = np.float(get_bsoup_data(bsoup_metadata, "ZENITH_ANGLE").string)
    azimuth_angle = np.float(get_bsoup_data(bsoup_metadata, "AZIMUTH_ANGLE").string)

    return (zenith_angle, azimuth_angle)




# Read from TOAR product's 'MTD_MSIL1C.xml' metadata file.
def get_proc_metadata(proc_metadata_filename, bands_to_process):

    with open(proc_metadata_filename, 'r') as fid:
        metadata = fid.read()
        bsoup_metadata = BeautifulSoup(metadata, "xml")

    # Pre-emptive initializations -- got 'em if we need 'em.
    num_bands_to_process = len(bands_to_process)
    
    metadata_band_attrs = map_imageband_to_metadataband_id_attribute(bands_to_process)
    if not metadata_band_attrs:
        err_msg = 'ERROR: selected band(s) not found'
        print(err_msg)
        return {}

    
    # Get the quantification value (scaling factor to convert true reflectance
    # to/from integer DN values.
    bsoup_quant_val = get_bsoup_data(bsoup_metadata, "QUANTIFICATION_VALUE")
    quant_val = np.int(bsoup_quant_val.string)

    # Get the processing baseline version.
    bsoup_baseline = get_bsoup_data(bsoup_metadata, "PROCESSING_BASELINE")
    baseline = Decimal(bsoup_baseline.string)

    # Get the spatial resolutions for all bands. These can
    # be pulled from the metadata.
    bsoup_resolution = bsoup_metadata.find_all('RESOLUTION')
    num_resolutions = len(bsoup_resolution)
    resolution_list = []
    bands_to_process_index = 0
    for i in range(num_resolutions):
        if bands_to_process_index == num_bands_to_process:
            break

        if bands_to_process[bands_to_process_index] == image_band_list[i]:
            res_val = np.int(bsoup_resolution[i].string)
            resolution_list.append(res_val)
            bands_to_process_index += 1

    # Get the radiometric offsets for each band in the
    # band list.
    rad_offset_data_list = []
    if baseline >= Decimal('4.00'):

        bsoup_rad_offsets = bsoup_metadata.find_all('RADIO_ADD_OFFSET')
        num_band_rad_offsets = len(bsoup_rad_offsets)
        bands_to_process_index = 0

        # Assumes/expects the # of bands to process will be less than
        # the number of bands with radiometric offset information.
        for i in range(num_band_rad_offsets):
            if bands_to_process_index == num_bands_to_process:
                break
            if bsoup_rad_offsets[i]['band_id'] == metadata_band_attrs[bands_to_process_index]:
                offset_val = np.int(bsoup_rad_offsets[bands_to_process_index].string)
                rad_offset_data_list.append(offset_val);
                bands_to_process_index += 1
    else:
        for band in bands_to_process:
            rad_offset_data_list.append(0)


    # Assemble the final processing metadata dictionary.
    metadata_dict = {}
    for i in range(num_bands_to_process):
        metadata_dict.update({bands_to_process[i]: \
            [quant_val, resolution_list[i], rad_offset_data_list[i]]})

    return metadata_dict


