import os
from pathlib import Path

import numpy as np
from osgeo import gdalconst

import cmac_gdal_io as io



def get_atmi_summary_stats(atmi_image_filename,
                           nodata_value):

    # Open the image dataset.
    atmi_image_data = io.read_image_from_file(atmi_image_filename)
    if atmi_image_data is None:
        err_msg = f'ERROR: cannot read image data from "{atmi_image_filename}"'
        print(err_msg)
        return None

    # Get the required summary statistics, taking into account the nodata
    # value.
    valid_pix = (atmi_image_data != nodata_value)
    min_val = np.min(atmi_image_data[valid_pix])
    max_val = np.max(atmi_image_data[valid_pix])
    avg_val = np.mean(atmi_image_data[valid_pix])
    stdev_val = np.std(atmi_image_data[valid_pix])

    return (min_val, max_val, avg_val, stdev_val)




def generate_summary_report(atmi_filename, nodata_value,
                            bands_list, atmi_elapsed_time_seconds,
                            correct_elapsed_time_seconds,
                            proc_elapsed_time_seconds,
                            num_negative_overcorrects_list,
                            num_below_one_overcorrects_list,
                            num_above_10k_overcorrects_list,
                            cmac_report_path, s2_image_basename,
                            cmac_proc_date, cmac_cal_model_version,
                            cmac_software_version):

    # Get the summary statistics for the Atm-I image.
    atmi_summary_stats = get_atmi_summary_stats(atmi_filename, nodata_value)
    if atmi_summary_stats is None:
        err_msg = f'ERROR: cannot retrieve summary Atm-I statistics'
        print(err_msg)
        return -1

    # Create the report file.
    if not Path(cmac_report_path).exists():
        err_msg = f'ERROR: CMAC output path "{cmac_report_path}" not found'
        print(err_msg)
        return -1

    report_filename = cmac_report_path + os.sep + s2_image_basename \
        + '_processing_summary.txt'

    with open(report_filename, 'wt') as out_fid:

        # Overall processing:
        out_line = f'PROCESSING SUMMARY FOR IMAGE TILE {s2_image_basename}'
        out_fid.write(out_line + '\n')
        out_fid.write('\n')
        out_line = f'DATE PROCESSED:             {cmac_proc_date}'
        out_fid.write(out_line + '\n')
        out_line = f'CALIBRATION MODEL VERSION:  {cmac_cal_model_version}'
        out_fid.write(out_line + '\n')
        out_line = f'CMAC SOFTWARE VERSION:      {cmac_software_version}'
        out_fid.write(out_line + '\n')
        out_line = f'TOTAL PROCESSING TIME:      {proc_elapsed_time_seconds:.2f} s'
        out_fid.write(out_line + '\n')
        out_fid.write('\n\n')

        # ATMI estimation.
        out_line = 'Atm-I ESTIMATION'
        out_fid.write(out_line + '\n')
        out_line = '---------------'
        out_fid.write(out_line + '\n')
        out_line = f'ESTIMATION TIME:            {atmi_elapsed_time_seconds:.2f} s'
        out_fid.write(out_line + '\n')
        out_line = f'Atm-I MINIMUM:              {atmi_summary_stats[0]}'
        out_fid.write(out_line + '\n')
        out_line = f'Atm-I MAXIMUM:              {atmi_summary_stats[1]}'
        out_fid.write(out_line + '\n')
        out_line = f'Atm-I MEAN:                 {atmi_summary_stats[2]:.2f}'
        out_fid.write(out_line + '\n')
        out_line = f'Atm-I STDEV:                {atmi_summary_stats[3]:.2f}'
        out_fid.write(out_line + '\n')
        out_fid.write('\n\n')

        # Correction summary info.
        out_line = 'CMAC CORRECTION SUMMARY, BY BAND:'
        out_fid.write(out_line + '\n')
        out_line = '--------------------------------'
        out_fid.write(out_line + '\n')
        out_line = f'CORRECTION TIME:            {correct_elapsed_time_seconds:.2f} s'
        out_fid.write(out_line + '\n')
        
        num_bands_corrected = len(bands_list)
        for i in range(num_bands_corrected):

            out_line = f'BAND {bands_list[i]} # OF PIXELS CORRECTING ' \
                + f'TO SR < 0:                         {num_negative_overcorrects_list[i]}'
            out_fid.write(out_line + '\n')
            out_line = f'BAND {bands_list[i]} # OF PIXELS CORRECTING ' \
                + f'TO SR < 1:                         {num_below_one_overcorrects_list[i]}'
            out_fid.write(out_line + '\n')
            out_line = f'BAND {bands_list[i]} # OF PIXELS CORRECTING ' \
                + f'TO SR > 10000:                     {num_above_10k_overcorrects_list[i]}'
            out_fid.write(out_line + '\n')
            out_fid.write('\n')

    return 0





