import json
import os
from pathlib import Path
import shutil

from cmac_metadata import map_band_index_to_imageband



# Allow for file structure on host
# running this application to be as
# follows:
# /<root_toa_folder>/<client_id>/<product_id>/TOA
# OR
# /<root_toa_folder>/<client_id>/<product_id>.SAFE

def is_SAFE_folder(folder_name):

    is_SAFE_folder = False
    if '.SAFE' in folder_name:
        is_SAFE_folder = True

    return is_SAFE_folder




def get_s2_tile_id_from_product_id(product_id):
    return product_id.split('_')[5]




def get_s2_datatake_datetime_from_product_id(product_id):
    return product_id.split('_')[2]




def construct_s2_image_basename(product_id):

    # Get the tile ID and datatake datetime from the product
    # ID.
    s2_tile_id = get_s2_tile_id_from_product_id(product_id)
    s2_datatake_datetime \
        = get_s2_datatake_datetime_from_product_id(product_id)

    return (s2_tile_id + '_' + s2_datatake_datetime)




def construct_parent_path(root_drive, user_id,
                          product_id, path_switch,
                          cmac_proc_date):

    # Only two paths possible: TOA or CMAC. If
    # the switch string isn't either of these,
    # that's an error.
    switch = path_switch.upper()
    if switch not in ['TOA', 'CMAC']:
        err_msg = f'ERROR: invalid path switch option ' \
            + f'"{path_switch}"'
        print(err_msg)
        return ''

    # If product_id contains .SAFE and we're constructing the
    # CMAC output path, strip off the .SAFE part. If we're
    # constructing the TOA input path, and the data is still
    # in the .SAFE "superfolder", we'll need to conserve that,
    # otherwise we'll need to add a 'TOA' child folder.
    is_SAFE_product =  is_SAFE_folder(product_id)
    if is_SAFE_product and switch == 'CMAC':
        product_id2 = product_id.split('.')[0]
    else:
        product_id2 = product_id

    # Start building.
    parent_path = root_drive + '/' + user_id + '/' \
        + product_id2 + '/3x3Average/30m'
    if switch == 'CMAC':
        parent_path = parent_path + '/' + 'CMAC_Surface' \
            + '/' + cmac_proc_date
    elif not is_SAFE_product:
        parent_path = parent_path + '/' + 'TOA'

    return parent_path




# Assumes/expects to find the 4 VNIR bands (RGB, 8A),
# MTD_TL.xml, and MTD_MSIL1C.xml to be "loose", in the
# folder <root_toa_folder>/<user_id>/<product_id>, if
# they are not contained in the .SAFE container folder.
def get_metadata_filename(parent_toa_path, metadata_type):

    if not Path(parent_toa_path).exists():
        err_msg = f'ERROR: "{parent_toa_path}" not found'
        print(err_msg)
        return ''

    metadata_filename = ''

    # The processing metadata folder is in the parent
    # folder. As of 03/23/2023, this logic will work with
    # Sentinel-2 .SAFE product folders.
    if metadata_type.upper() not in ['PROCESSING', 'TILE']:
        err_msg = f'ERROR: invalid metadata file type "{metadata_type}"'
        print(err_msg)
        return ''

    elif metadata_type == 'PROCESSING':
        metadata_filename = parent_toa_path + '/' + 'MTD_MSIL1C.xml'

    else:
        # Go after the tile metadata folder "MTD_TL.xml'
        product_path = ''

        # Is the image data folder the product .SAFE folder?
        if is_SAFE_folder(parent_toa_path):

            # Build the path to the folder.
            granule_path = parent_toa_path + '/' + 'GRANULE'
            if not Path(granule_path).exists():
                err_msg = f'ERROR: "GRANULE" child folder not found'
                print(err_msg)
                return ''
        
            # The processing metadata folder is in the 
            # Get the contents of the 'GRANULE' subfolder.
            granule_dir_list = [str(d) for d in Path(granule_path).glob('L1C*')
                              if d.is_dir()]
            num_granule_dirs = len(granule_dir_list)
            if num_granule_dirs != 1:
                err_msg = 'ERROR: cannot reliably determine final "tile" ' \
                    + 'metadata file name under {granule_path}'
                print(err_msg)
                return ''

            product_path = granule_dir_list[0]
        else:
            product_path = parent_toa_path

        metadata_filename = product_path + '/' + 'MTD_TL.xml'

    # Done.
    return metadata_filename



# For .SAFE folder location, assumes/expects the
# current S2 file location/naming conventions apply.
def get_toa_image_path(parent_toa_path):

    # Make sure the parent TOA folder exists.
    if not Path(parent_toa_path).exists():
        err_msg = f'ERROR: "{parent_toa_path}" not found'
        print(err_msg)
        return ''

    # Is the parent TOA path the .SAFE folder?
    image_path = ''
    if is_SAFE_folder(parent_toa_path):

        # Build the path to the IMG_DATA folder.
        granule_path = parent_toa_path + '/' + 'GRANULE'
        if not Path(granule_path).exists():
            err_msg = f'ERROR: "GRANULE" child folder not found'
            print(err_msg)
            return ''
        
        # Get the contents of the 'GRANULE' subfolder.
        granule_dir_list = [str(d) for d in Path(granule_path).glob('L1C*')
                          if d.is_dir()]
        num_granule_dirs = len(granule_dir_list)
        if num_granule_dirs != 1:
            err_msg = 'ERROR: cannot reliably determine final "tile" ' \
                + 'metadata file name under "{granule_path}"'
            print(err_msg)
            return ''
        image_path = granule_dir_list[0] + '/' + 'IMG_DATA'
    else:
        # Assumes path structure is /<root_toa_folder>/<user_id>/<product_id>/TOA
        image_path = parent_toa_path

    return image_path




# Function to construct partial or total absolute paths for image
# data files.
def construct_image_filenames(parent_path, s2_image_basename,
                              prefix, suffix, bands_to_process_list,
                              file_ext):

    # Initializations.
    image_filenames_list = []
    num_bands_to_process = len(bands_to_process_list)
    num_filenames_constructed = 0

    for band in bands_to_process_list:

        curr_filename = parent_path + '/'
        if prefix != '':
            curr_filename = curr_filename + prefix + '_'
        curr_filename = curr_filename + s2_image_basename

        if suffix != '':
            curr_filename = curr_filename + '_' + suffix
        curr_filename = curr_filename + '_' + band \
            + '.' + file_ext

        # Add this filename to the list.
        image_filenames_list.append(curr_filename)
        num_filenames_constructed += 1

    # Did we create all the filenames we needed to?
    if num_filenames_constructed != num_bands_to_process:
        err_msg = f'ERROR: {num_filenames_constructed} image filenames ' \
            + f'constructed, but {num_bands_to_process} are required'
        print(err_msg)
        return []

    return image_filenames_list



def clean_cmac_directory(cmac_image_path):

    # Get the list of filenames currently in the directory.
    path = Path(cmac_image_path)
    if not path.exists():
        err_msg = f'{cmac_image_path} not found, cannot clean'
        print(err_msg)
        return
   
    # Do a "clean sweep", delete all the files and the folder
    # itself.
    contents_list = path.glob('*')
    for item in contents_list:
        fpath = Path(item)
        if fpath.is_file():
            fpath.unlink()
        elif fpath.is_dir():
            shutil.rmtree(fpath)




# Run this at the tail end, after the report has
# been generated.
def clean_cmac_gtiffs(cmac_image_path):
    path = Path(cmac_image_path)
    if not path.exists():
        err_msg = f'{cmac_image_path} not found, cannot clean'
        print(err_msg)
        return

    # Selectively remove the GTiff image files.
    GTiff_file_list = path.glob('*.tif')
    for item in GTiff_file_list:
        fpath = Path(item)
        if fpath.is_file():
            fpath.unlink()



def get_band_ids_list(band_indices_list):

    num_indices_in_list = len(band_indices_list)

    band_ids_list = []
    num_bands_found = 0
    for i in range(num_indices_in_list):
        curr_id = map_band_index_to_imageband(band_indices_list[i])
        if curr_id != '':
            num_bands_found += 1
            band_ids_list.append(curr_id)

    if num_bands_found != num_indices_in_list:
        err_msg = f'ERROR: {num_bands_found} were found for list, ' \
            + f'{num_indices_in_list} bands are required'
        print(err_msg)
        return []

    return band_ids_list




# Polynomial evaluation using Horner's algorithm.
def eval_horner_poly(x, coeffs_list, num_coeffs):

    # Start with the highest power coefficient and work
    # backwords. Assumes polynomial coefficients are ordered
    # from const to highest powered term.
    poly_value = coeffs_list[num_coeffs-1];
    #print(poly_value)

    # End value in a python(range) interval is NOT inclusive --
    # need to add/subtract one to get ALL elements considered
    for i in range(num_coeffs-2, -1, -1):
        poly_value *= x
        poly_value += coeffs_list[i]
        #print(poly_value)

    return poly_value




def main():
    root_drive = 'D:/Tim/CMACProcImagery'
    client_id = 'ARSI'
    product_id = 'S2A_MSIL1C_20180806T171901_N0206_R012_T14TPP_20180806T212121'
    safe_id = 'S2A_MSIL1C_20180806T171901_N0206_R012_T14TPP_20180806T212121.SAFE'

    cmac_proc_date = '20230325'

    s2_image_basename = construct_s2_image_basename(product_id)
    print(f'S2 IMAGE BASENAME:    {s2_image_basename}')

    parent_toa_path = construct_parent_path(root_drive, client_id, product_id,
                                            'TOA', cmac_proc_date)
    print(f'PARENT TOA PATH:      {parent_toa_path}')

    parent_cmac_path = construct_parent_path(root_drive, client_id, product_id,
                                             'CMAC', cmac_proc_date)
    print(f'PARENT CMAC PATH:     {parent_cmac_path}')


    parent_toa_path2 = construct_parent_path(root_drive, client_id, safe_id,
                                             'TOA', cmac_proc_date)
    print(f'PARENT TOA PATH:      {parent_toa_path2}')

    parent_cmac_path2 = construct_parent_path(root_drive, client_id, safe_id,
                                              'CMAC', cmac_proc_date)
    print(f'PARENT CMAC PATH:     {parent_cmac_path2}')


    # Test Horner's polynomial eval.
    coeffs_list = [1, 2, 1]
    x = 2
    poly_val = eval_horner_poly(x, coeffs_list, len(coeffs_list))
    print(f'POLY COEFFS:  {coeffs_list}')
    print(f'POLYNOMIAL EVALUATION, x={x}:  {poly_val}')

    coeffs2_list = [9, 6, 1] # (x+3)^2 -> 9 + 6x + x^2
    poly_val2 = eval_horner_poly(x, coeffs2_list, len(coeffs2_list))
    print(f'POLY COEFFS:  {coeffs2_list}')
    print(f'POLYNOMIAL EVALUATION, x={x}:  {poly_val2}')



if __name__ == '__main__':
    main()

        
